﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{   
    public class CodeEditorBase : Form
    {

        #region Constants

        internal const int REFRESH_TIMER_INTERVAL = 500;
        internal const int PERF_THRESHOLD = 1000;

        internal static readonly string PATH = Application.StartupPath.Replace("\\TestApp\\bin\\Debug", string.Empty);
        internal static readonly string PATH_GRAMMAR_DEF = PATH + @"\GrammarDefinitions";
        internal static readonly string PATH_SOURCE = PATH + @"\SourceFiles";

        static internal readonly Encoding TextEncoding = Encoding.ASCII;

        #endregion Constants

        #region Member variables

        protected System.ComponentModel.IContainer components = null;

        protected SplitContainer splitContainer1;
        protected RichTextBox txtCode;
        protected MenuStrip menuStrip1;
        protected ListBox lstOutput;
        protected StatusStrip statusStrip1;

        protected ToolStripMenuItem statusErrors = new ToolStripMenuItem();
        protected ToolStripMenuItem statusTime = new ToolStripMenuItem();
        protected ToolStripMenuItem statusLineNumber = new ToolStripMenuItem();
        protected ToolStripMenuItem statusColumn = new ToolStripMenuItem();
        protected ToolStripMenuItem fileToolStripMenuItem;
        protected ToolStripMenuItem openToolStripMenuItem;
        protected ToolStripMenuItem closeToolStripMenuItem;
        protected ToolStripSeparator toolStripSeparator1;
        protected ToolStripMenuItem saveToolStripMenuItem;
        protected ToolStripMenuItem saveAsToolStripMenuItem;
        protected ToolStripSeparator toolStripSeparator2;
        protected ToolStripMenuItem exitToolStripMenuItem;
        protected ToolStripMenuItem editToolStripMenuItem;
        protected ToolStripMenuItem findToolStripMenuItem;
        protected ToolStripMenuItem replaceToolStripMenuItem;
        protected ToolStripSeparator toolStripSeparator3;
        protected ToolStripMenuItem copyToolStripMenuItem;
        protected ToolStripMenuItem pasteToolStripMenuItem;
        protected ToolStripMenuItem viewToolStripMenuItem;
        protected ToolStripMenuItem outputToolStripMenuItem;
        protected ToolStripMenuItem ruleTreeToolStripMenuItem;
        protected ToolStripMenuItem statusCharacter = new ToolStripMenuItem();

        protected ErrorTokenToolTip toolTip = null;

        protected bool modified = false;
        protected bool suspended = false;
        protected string filename = null;
        private ToolStripSeparator toolStripSeparator4;
        protected frmFind findForm = null;

        private Regex regexIdentifier = new Regex( GrammarImporter.IdentifierRegexPattern );
        private Regex regexBackToDelimiter = new Regex( "[^a-zA-Z0-9_@]|^[\\s\t\n]", RegexOptions.RightToLeft );

        #endregion Member variables

        #region Constructor
        public CodeEditorBase()
        {
            this.InitializeComponent();
        }
        #endregion Constructor

        #region InitializeComponent
        private void InitializeComponent()
        {
            this.splitContainer1 = new System.Windows.Forms.SplitContainer();
            this.txtCode = new System.Windows.Forms.RichTextBox();
            this.menuStrip1 = new System.Windows.Forms.MenuStrip();
            this.fileToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.openToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.closeToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.toolStripSeparator1 = new System.Windows.Forms.ToolStripSeparator();
            this.saveToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.saveAsToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.toolStripSeparator2 = new System.Windows.Forms.ToolStripSeparator();
            this.exitToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.editToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.findToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.replaceToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.toolStripSeparator3 = new System.Windows.Forms.ToolStripSeparator();
            this.copyToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.pasteToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.viewToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.outputToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.ruleTreeToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.lstOutput = new System.Windows.Forms.ListBox();
            this.statusStrip1 = new System.Windows.Forms.StatusStrip();
            this.toolStripSeparator4 = new System.Windows.Forms.ToolStripSeparator();
            this.splitContainer1.Panel1.SuspendLayout();
            this.splitContainer1.Panel2.SuspendLayout();
            this.splitContainer1.SuspendLayout();
            this.menuStrip1.SuspendLayout();
            this.SuspendLayout();
            // 
            // splitContainer1
            // 
            this.splitContainer1.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
                        | System.Windows.Forms.AnchorStyles.Left)
                        | System.Windows.Forms.AnchorStyles.Right)));
            this.splitContainer1.Location = new System.Drawing.Point(0, 0);
            this.splitContainer1.Name = "splitContainer1";
            this.splitContainer1.Orientation = System.Windows.Forms.Orientation.Horizontal;
            // 
            // splitContainer1.Panel1
            // 
            this.splitContainer1.Panel1.Controls.Add(this.txtCode);
            this.splitContainer1.Panel1.Controls.Add(this.menuStrip1);
            // 
            // splitContainer1.Panel2
            // 
            this.splitContainer1.Panel2.Controls.Add(this.lstOutput);
            this.splitContainer1.Size = new System.Drawing.Size(683, 387);
            this.splitContainer1.SplitterDistance = 301;
            this.splitContainer1.TabIndex = 7;
            // 
            // txtCode
            // 
            this.txtCode.Dock = System.Windows.Forms.DockStyle.Fill;
            this.txtCode.Location = new System.Drawing.Point(0, 24);
            this.txtCode.Name = "txtCode";
            this.txtCode.ShowSelectionMargin = true;
            this.txtCode.Size = new System.Drawing.Size(683, 277);
            this.txtCode.TabIndex = 2;
            this.txtCode.Text = "";
            this.txtCode.WordWrap = false;
            // 
            // menuStrip1
            // 
            this.menuStrip1.Items.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.fileToolStripMenuItem,
            this.editToolStripMenuItem,
            this.viewToolStripMenuItem});
            this.menuStrip1.Location = new System.Drawing.Point(0, 0);
            this.menuStrip1.Name = "menuStrip1";
            this.menuStrip1.Size = new System.Drawing.Size(683, 24);
            this.menuStrip1.TabIndex = 1;
            this.menuStrip1.Text = "menuStrip1";
            // 
            // fileToolStripMenuItem
            // 
            this.fileToolStripMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.openToolStripMenuItem,
            this.closeToolStripMenuItem,
            this.toolStripSeparator1,
            this.saveToolStripMenuItem,
            this.saveAsToolStripMenuItem,
            this.toolStripSeparator2,
            this.exitToolStripMenuItem});
            this.fileToolStripMenuItem.Name = "fileToolStripMenuItem";
            this.fileToolStripMenuItem.Size = new System.Drawing.Size(37, 20);
            this.fileToolStripMenuItem.Text = "&File";
            // 
            // openToolStripMenuItem
            // 
            this.openToolStripMenuItem.Name = "openToolStripMenuItem";
            this.openToolStripMenuItem.Size = new System.Drawing.Size(114, 22);
            this.openToolStripMenuItem.Text = "&Open";
            // 
            // closeToolStripMenuItem
            // 
            this.closeToolStripMenuItem.Name = "closeToolStripMenuItem";
            this.closeToolStripMenuItem.Size = new System.Drawing.Size(114, 22);
            this.closeToolStripMenuItem.Text = "&Close";
            // 
            // toolStripSeparator1
            // 
            this.toolStripSeparator1.Name = "toolStripSeparator1";
            this.toolStripSeparator1.Size = new System.Drawing.Size(111, 6);
            // 
            // saveToolStripMenuItem
            // 
            this.saveToolStripMenuItem.Name = "saveToolStripMenuItem";
            this.saveToolStripMenuItem.Size = new System.Drawing.Size(114, 22);
            this.saveToolStripMenuItem.Text = "&Save";
            // 
            // saveAsToolStripMenuItem
            // 
            this.saveAsToolStripMenuItem.Name = "saveAsToolStripMenuItem";
            this.saveAsToolStripMenuItem.Size = new System.Drawing.Size(114, 22);
            this.saveAsToolStripMenuItem.Text = "Save &As";
            // 
            // toolStripSeparator2
            // 
            this.toolStripSeparator2.Name = "toolStripSeparator2";
            this.toolStripSeparator2.Size = new System.Drawing.Size(111, 6);
            // 
            // exitToolStripMenuItem
            // 
            this.exitToolStripMenuItem.Name = "exitToolStripMenuItem";
            this.exitToolStripMenuItem.Size = new System.Drawing.Size(114, 22);
            this.exitToolStripMenuItem.Text = "E&xit";
            // 
            // editToolStripMenuItem
            // 
            this.editToolStripMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.findToolStripMenuItem,
            this.replaceToolStripMenuItem,
            this.toolStripSeparator3,
            this.copyToolStripMenuItem,
            this.pasteToolStripMenuItem});
            this.editToolStripMenuItem.Name = "editToolStripMenuItem";
            this.editToolStripMenuItem.Size = new System.Drawing.Size(39, 20);
            this.editToolStripMenuItem.Text = "&Edit";
            // 
            // findToolStripMenuItem
            // 
            this.findToolStripMenuItem.Name = "findToolStripMenuItem";
            this.findToolStripMenuItem.Size = new System.Drawing.Size(115, 22);
            this.findToolStripMenuItem.Text = "F&ind";
            // 
            // replaceToolStripMenuItem
            // 
            this.replaceToolStripMenuItem.Name = "replaceToolStripMenuItem";
            this.replaceToolStripMenuItem.Size = new System.Drawing.Size(115, 22);
            this.replaceToolStripMenuItem.Text = "&Replace";
            // 
            // toolStripSeparator3
            // 
            this.toolStripSeparator3.Name = "toolStripSeparator3";
            this.toolStripSeparator3.Size = new System.Drawing.Size(112, 6);
            // 
            // copyToolStripMenuItem
            // 
            this.copyToolStripMenuItem.Name = "copyToolStripMenuItem";
            this.copyToolStripMenuItem.Size = new System.Drawing.Size(115, 22);
            this.copyToolStripMenuItem.Text = "&Copy";
            // 
            // pasteToolStripMenuItem
            // 
            this.pasteToolStripMenuItem.Name = "pasteToolStripMenuItem";
            this.pasteToolStripMenuItem.Size = new System.Drawing.Size(115, 22);
            this.pasteToolStripMenuItem.Text = "&Paste";
            // 
            // viewToolStripMenuItem
            // 
            this.viewToolStripMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.outputToolStripMenuItem,
            this.toolStripSeparator4,
            this.ruleTreeToolStripMenuItem});
            this.viewToolStripMenuItem.Name = "viewToolStripMenuItem";
            this.viewToolStripMenuItem.Size = new System.Drawing.Size(44, 20);
            this.viewToolStripMenuItem.Text = "&View";
            // 
            // outputToolStripMenuItem
            // 
            this.outputToolStripMenuItem.Name = "outputToolStripMenuItem";
            this.outputToolStripMenuItem.Size = new System.Drawing.Size(152, 22);
            this.outputToolStripMenuItem.Text = "O&utput";
            // 
            // ruleTreeToolStripMenuItem
            // 
            this.ruleTreeToolStripMenuItem.Name = "ruleTreeToolStripMenuItem";
            this.ruleTreeToolStripMenuItem.Size = new System.Drawing.Size(152, 22);
            this.ruleTreeToolStripMenuItem.Text = "&Rule Tree";
            // 
            // lstOutput
            // 
            this.lstOutput.Dock = System.Windows.Forms.DockStyle.Fill;
            this.lstOutput.FormattingEnabled = true;
            this.lstOutput.IntegralHeight = false;
            this.lstOutput.Location = new System.Drawing.Point(0, 0);
            this.lstOutput.Name = "lstOutput";
            this.lstOutput.Size = new System.Drawing.Size(683, 82);
            this.lstOutput.TabIndex = 0;
            // 
            // statusStrip1
            // 
            this.statusStrip1.LayoutStyle = System.Windows.Forms.ToolStripLayoutStyle.HorizontalStackWithOverflow;
            this.statusStrip1.Location = new System.Drawing.Point(0, 390);
            this.statusStrip1.Name = "statusStrip1";
            this.statusStrip1.Size = new System.Drawing.Size(684, 22);
            this.statusStrip1.TabIndex = 8;
            this.statusStrip1.Text = "statusStrip1";
            // 
            // toolStripSeparator4
            // 
            this.toolStripSeparator4.Name = "toolStripSeparator4";
            this.toolStripSeparator4.Size = new System.Drawing.Size(149, 6);
            // 
            // CodeEditorBase
            // 
            this.ClientSize = new System.Drawing.Size(684, 412);
            this.Controls.Add(this.splitContainer1);
            this.Controls.Add(this.statusStrip1);
            this.Name = "CodeEditorBase";
            this.Text = "Code editor";
            this.splitContainer1.Panel1.ResumeLayout(false);
            this.splitContainer1.Panel1.PerformLayout();
            this.splitContainer1.Panel2.ResumeLayout(false);
            this.splitContainer1.ResumeLayout(false);
            this.menuStrip1.ResumeLayout(false);
            this.menuStrip1.PerformLayout();
            this.ResumeLayout(false);
            this.PerformLayout();

        }
        #endregion InitializeComponent

        #region InitializeUI
        protected virtual void InitializeUI()
        {
            this.toolTip = new ErrorTokenToolTip( this.txtCode );
            this.saveToolStripMenuItem.Enabled = false;
            this.closeToolStripMenuItem.Enabled = false;

            this.txtCode.HideSelection = false;
            this.txtCode.TextChanged += new EventHandler( this.txtCode_TextChanged );
            this.txtCode.SelectionChanged += new EventHandler(this.txtCode_SelectionChanged);
            this.txtCode.Font = new Font( "Lucida Console", 10f );
            this.txtCode.AcceptsTab = true;            
            this.txtCode_TextChanged( this.txtCode, EventArgs.Empty );

            this.lstOutput.MouseDoubleClick += new MouseEventHandler( this.lstOutput_MouseDoubleClick );
            EventHandler contextMenuClickHandler = new EventHandler(this.OnOutputMenuItemClicked);
            ContextMenu cm = new ContextMenu();
            MenuItem mi = null;
            mi = cm.MenuItems.Add( "Copy", contextMenuClickHandler );
            mi.Name = "copy";
            mi = cm.MenuItems.Add( "Clear", contextMenuClickHandler );
            mi.Name = "clear";
            this.lstOutput.ContextMenu = cm;

            EventHandler menuStripClickHandler = new EventHandler(this.OnMenuStripMenuItemClick);
            this.saveToolStripMenuItem.Click += menuStripClickHandler;
            this.saveAsToolStripMenuItem.Click += menuStripClickHandler;
            this.exitToolStripMenuItem.Click += menuStripClickHandler;
            this.outputToolStripMenuItem.Click += menuStripClickHandler;
            this.ruleTreeToolStripMenuItem.Click += menuStripClickHandler;
            this.openToolStripMenuItem.Click += menuStripClickHandler;
            this.closeToolStripMenuItem.Click += menuStripClickHandler;
            this.findToolStripMenuItem.Click += menuStripClickHandler;
            this.replaceToolStripMenuItem.Click += menuStripClickHandler;
            this.copyToolStripMenuItem.Click += menuStripClickHandler;
            this.pasteToolStripMenuItem.Click += menuStripClickHandler;

            this.outputToolStripMenuItem.Checked = false;
            this.splitContainer1.Panel2Collapsed = true;

            this.saveToolStripMenuItem.ShortcutKeys = (Keys.Control | Keys.S);
            this.openToolStripMenuItem.ShortcutKeys = (Keys.Control | Keys.O);
            this.findToolStripMenuItem.ShortcutKeys = (Keys.Control | Keys.F);
            this.replaceToolStripMenuItem.ShortcutKeys = (Keys.Control | Keys.H);
            this.copyToolStripMenuItem.ShortcutKeys = (Keys.Control | Keys.Insert);
            this.pasteToolStripMenuItem.ShortcutKeys = (Keys.Shift | Keys.Insert);

            this.statusErrors.AutoToolTip = false;            
            this.statusLineNumber.Alignment = ToolStripItemAlignment.Right;
            this.statusColumn.Alignment = ToolStripItemAlignment.Right;
            this.statusCharacter.Alignment = ToolStripItemAlignment.Right;
            this.statusStrip1.Items.Add( this.statusErrors ); 
            this.statusStrip1.Items.Add( this.statusTime ); 
            this.statusStrip1.Items.Add( this.statusCharacter );
            this.statusStrip1.Items.Add( this.statusColumn );
            this.statusStrip1.Items.Add( this.statusLineNumber );

            this.toolTip = new ErrorTokenToolTip( this.txtCode );

            this.SetModified( false );
            this.txtCode.Enabled = false;
        }
        #endregion InitializeUI

        #region RenameMenuItems
        protected void RenameMenuItems( MenuStrip menuStrip )
        {
            List<ToolStripMenuItem> list = new List<ToolStripMenuItem>();
            foreach( ToolStripMenuItem mi in menuStrip.Items )
            {
                string name = mi.Name.Replace("ToolStripMenuItem", string.Empty);
                mi.Name = name;

                CodeEditorBase.GetMenuItems( mi.DropDownItems, list );
            }

            foreach ( ToolStripItem item in list )
            {
                string name = item.Name.Replace("ToolStripMenuItem", string.Empty);
                item.Name = name;
                Debug.WriteLine( name );
            }
        }

        static private void GetMenuItems( ToolStripItemCollection items, List<ToolStripMenuItem> list )
        {
            foreach ( ToolStripItem item in items )
            {
                ToolStripMenuItem menuItem = item as ToolStripMenuItem;
                if ( menuItem == null )
                    continue;

                list.Add( menuItem );

                if ( menuItem.HasDropDownItems )
                    CodeEditorBase.GetMenuItems( menuItem.DropDownItems, list );
            }
        }
        #endregion RenameMenuItems

        #region GetLineNumber
        protected virtual int GetLineNumber( string input, out int column )
        {
            throw new NotImplementedException();
        }
        #endregion GetLineNumber

        #region NewLine
        protected virtual string NewLine
        {
            get { throw new NotImplementedException(); }
        }
        #endregion NewLine

        #region Grammar
        protected virtual Grammar Grammar { get { throw new NotImplementedException(); } }
        #endregion Grammar

        #region TextHighlighter
        protected virtual TextHighlighter TextHighlighter { get { throw new NotImplementedException(); } }
        #endregion TextHighlighter

        #region InitializeOpenFileDialog / InitializeSaveFileDialog
        protected virtual void InitializeOpenFileDialog( OpenFileDialog dialog )
        {
            throw new NotImplementedException();
        }

        protected virtual void InitializeSaveFileDialog( SaveFileDialog dialog )
        {
            throw new NotImplementedException();
        }
        #endregion InitializeOpenFileDialog / InitializeSaveFileDialog

        #region txtCode_TextChanged
        private void txtCode_TextChanged(object sender, EventArgs e)
        {
            if ( this.suspended == false )
            {
                this.OnTextBoxTextChanged();

                this.SetModified( true );

                bool enabled = this.txtCode.Text.Length > 0;
                this.saveToolStripMenuItem.Enabled = enabled;
                this.closeToolStripMenuItem.Enabled = enabled;
                
                if ( enabled == false )
                {
                    this.ClearResults();
                }
            }
        }
        
        protected virtual void OnTextBoxTextChanged(){}

        #endregion txtCode_TextChanged

        #region txtCode_SelectionChanged
        private void txtCode_SelectionChanged(object sender, EventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;
            int selectionStart = textBox.SelectionStart;
            this.copyToolStripMenuItem.Enabled = textBox.SelectionLength > 0;

            string text = textBox.Text;

            selectionStart = Math.Max( selectionStart, 0 );
            string s = text.Substring( 0, selectionStart );
            int column = 1;
            int line = this.GetLineNumber( s, out column );

            this.statusLineNumber.Text = string.Format( "Ln {0}", line );
            this.statusColumn.Text = string.Format( "Col {0}", column );
            this.statusCharacter.Text = string.Format( "Ch {0}", selectionStart );
        }
        #endregion txtCode_SelectionChanged

        #region SetModified
        protected void SetModified( bool value )
        {
            this.modified = value;

            this.saveToolStripMenuItem.Enabled = this.modified;
            this.saveAsToolStripMenuItem.Enabled = this.modified;
        }
        #endregion SetModified

        #region SetTimeElapsed
        protected void SetTimeElapsed( int value )
        {
            if ( value < 0 )
            {
                this.statusTime.Visible = false;
                return;
            }

            this.statusTime.Text = string.Format( "{0} msec.", value );
            this.statusTime.ForeColor = value > frmMain.PERF_THRESHOLD ? Color.Red : Color.Empty;
        }
        #endregion SetTimeElapsed

        #region LogResults
        protected void LogResults( IEnumerable<IErrorToken> errors )
        {
            this.lstOutput.Items.Clear();

            this.ruleTreeToolStripMenuItem.Enabled = true;

            string success = "0 errors.";
            int errorCount = 0;

            if ( errors != null )
            {
                foreach ( IErrorToken error in errors )
                {
                    errorCount++;
                    this.lstOutput.Items.Add( error );
                }
            }

            if ( errorCount == 0 )
            {
                this.statusErrors.ToolTipText = string.Empty;
                this.statusErrors.Text = success;
                this.statusErrors.Image = null;
            }
            else
            {
                this.statusErrors.Text = string.Format( "{0} error(s).", errorCount );
                this.statusErrors.Image = SystemIcons.Error.ToBitmap();
            }

        }

        protected void ClearResults()
        {
            this.lstOutput.Items.Clear();
            this.statusErrors.Text = string.Empty;
            this.statusErrors.Image = null;
            this.statusTime.Text = string.Empty;
        }
        #endregion LogResults

        #region PromptSave
        private bool PromptSave()
        {
            if ( this.modified )
            {
                if ( string.IsNullOrEmpty(this.filename) && this.txtCode.TextLength == 0 )
                    return true;

                DialogResult result = MessageBox.Show( "Contents have been modified. would you like to save?", "Save", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question );
                switch ( result )
                {
                    case DialogResult.Cancel:
                        return false;

                    case DialogResult.Yes:
                        this.OnMenuStripMenuItemClick( this.saveToolStripMenuItem, EventArgs.Empty );
                        break;
                }
            }

            return true;
        }
        #endregion PromptSave

        #region OnMenuStripMenuItemClick
        private void OnMenuStripMenuItemClick(object sender, EventArgs e)
        {
            ToolStripMenuItem mi = sender as ToolStripMenuItem;
            this.OnMenuStripMenuItemClick( mi );
        }

        protected bool OnMenuStripMenuItemClick(ToolStripMenuItem mi)
        {        
            string name = mi.Name;

            OpenFileDialog ofd = null;
            frmRuleTree ruleTree = null;
            bool handled = false;

            switch ( name )
            {
                case "save":

                    if ( string.IsNullOrEmpty(this.filename) )
                    {
                        this.OnMenuStripMenuItemClick( this.saveAsToolStripMenuItem, EventArgs.Empty );
                        break;
                    }

                    try
                    {
                        this.Cursor = Cursors.WaitCursor;
                        string fileText = this.txtCode.Text.Replace( this.NewLine, Environment.NewLine );
                        File.WriteAllText( this.filename, fileText, frmMain.TextEncoding );
                        this.statusErrors.Image = null;
                        this.statusErrors.Text = "File saved";
                        this.SetModified( false );
                    }
                    finally
                    {
                        this.Cursor = Cursors.Default;
                    }
                    handled = true;
                    break;

                case "saveAs":
                    SaveFileDialog sfd = new SaveFileDialog();
                    this.InitializeSaveFileDialog( sfd );
                    if ( sfd.ShowDialog(this) == DialogResult.OK )
                    {
                        string fileText = this.txtCode.Text.Replace( this.NewLine, Environment.NewLine );
                        File.WriteAllText( sfd.FileName, fileText, frmMain.TextEncoding );

                        if ( string.IsNullOrEmpty(this.filename) )
                            this.filename = sfd.FileName;

                        this.SetModified( false );
                    }

                    handled = true;
                    break;

                case "open":

                    if ( this.PromptSave() == false )
                        break;
                    
                    ofd = new OpenFileDialog();
                    this.InitializeOpenFileDialog( ofd );
                    if ( ofd.ShowDialog(this) == DialogResult.OK && File.Exists(ofd.FileName) )
                    {
                        Cursor.Current = Cursors.WaitCursor;

                        this.txtCode.Enabled = true;

                        this.filename = ofd.FileName;
                        
                        string s = File.ReadAllText( this.filename, frmMain.TextEncoding );                        
                        
                        try
                        {
                            this.TextHighlighter.Suspend();
                            this.txtCode.Text = s.Replace( Environment.NewLine, this.NewLine );
                        }
                        finally
                        {
                            this.TextHighlighter.Resume();
                        }

                        this.SetModified( false );
                        this.closeToolStripMenuItem.Enabled = true;

                        this.ClearResults();

                        Cursor.Current = Cursors.Default;
                    }

                    handled = true;
                    break;

                case "close":

                    if ( this.PromptSave() == false )
                    {
                        handled = true;
                        break;
                    }

                    this.filename = null;
                    this.txtCode.Clear();
                    this.SetModified( false );
                    this.closeToolStripMenuItem.Enabled = false;
                    handled = true;
                    break;
                    
                case "output":
                    mi.Checked = (mi.Checked == false);
                    this.splitContainer1.Panel2Collapsed = (mi.Checked == false);
                    handled = true;
                    break;

                case "ruleTree":
                    ruleTree = new frmRuleTree();
                    ruleTree.ShowMe( this, this.Grammar );
                    handled = true;
                    break;

                case "copy":
                    try { Clipboard.SetText( this.txtCode.SelectedText ); }
                    catch{}
                    handled = true;
                    break;

                case "paste":
                    string clipboardText = Clipboard.GetText();
                    if ( string.IsNullOrEmpty(clipboardText) == false )
                        this.txtCode.SelectedText = clipboardText;
                    handled = true;
                    break;

                case "find":
                case "replace":
                    
                    if ( this.findForm == null )
                    {
                        this.findForm = new frmFind();
                        this.findForm.Find += new FindHandler(this.findForm_Find);
                        this.findForm.Replace += new ReplaceHandler(this.findForm_Replace);
                        this.findForm.ReplaceAll += new ReplaceAllHandler(this.findForm_ReplaceAll);
                    }
                    else
                        this.findForm.Visible = true;

                    bool replace = (name == "replace");

                    string findText = this.txtCode.SelectedText;
                    if ( findText.Length == 0 && Clipboard.ContainsText() )
                        findText = Clipboard.GetText();

                    int selStart = this.txtCode.SelectionStart;
                    if ( replace == false )
                        selStart += this.txtCode.SelectionLength;

                    this.findForm.ShowMe( this, this.txtCode.Text, findText, replace, selStart );

                    handled = true;
                    break;

                case "exit":
                    this.Close();
                    handled = true;
                    break;
            }

            return handled;
            
        }

        #endregion OnMenuStripMenuItemClick

        #region OnOutputMenuItemClicked
        private void OnOutputMenuItemClicked(object sender, EventArgs e)
        {
            MenuItem mi = sender as MenuItem;

            switch ( mi.Name )
            {
                case "clear":
                    this.lstOutput.Items.Clear();
                    break;

                case "copy":

                    StringBuilder sb = new StringBuilder();
                    
                    foreach ( object o in this.lstOutput.Items )
                    {
                        sb.AppendLine( o.ToString() );
                    }

                    Clipboard.SetText( sb.ToString() );
                    
                    break;

            }
        }
        #endregion OnOutputMenuItemClicked

        #region lstOutput_MouseDoubleClick
        private void lstOutput_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            int index = this.lstOutput.IndexFromPoint( e.Location );
            IErrorToken error = index >= 0 ? this.lstOutput.Items[index] as IErrorToken : null;

            if ( error != null )
            {
                int charPos = error.CharacterPosition;
                this.txtCode.Select( charPos, 0 );
                this.txtCode.Select();
            }
        }
        #endregion lstOutput_MouseDoubleClick

        #region findForm_Find
        private void findForm_Find( object sender, FindEventArgs e )
        {
            if ( e.Success == false )
            {
                string message = string.Format("'{0}' not found.", e.Text);
                MessageBox.Show( message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information ); 
            }
            else
            {
                this.txtCode.SelectionStart = e.CharPos;
                this.txtCode.SelectionLength = e.Text.Length;
            }
        }
        #endregion findForm_Find

        #region findForm_Replace
        private void findForm_Replace(object sender, ReplaceEventArgs e)
        {
            this.txtCode.SelectionStart = e.CharPos;
            this.txtCode.SelectionLength = e.FindText.Length;
            this.txtCode.SelectedText = e.ReplaceText;
        }
        #endregion findForm_Replace

        #region findForm_ReplaceAll
        private void findForm_ReplaceAll(object sender, ReplaceEventArgsBase e)
        {
            this.txtCode.Text = this.txtCode.Text.Replace( e.FindText, e.ReplaceText );
        }
        #endregion findForm_ReplaceAll

        #region OnKeyDown
        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);

            if ( e.KeyData == Keys.F3 && this.findForm != null )
                e.Handled = this.findForm.FindNext();
        }
        #endregion OnKeyDown

        #region OnClosing
        protected override void OnClosing(CancelEventArgs e)
        {
            base.OnClosing(e);

            if ( this.PromptSave() == false )
                e.Cancel = true;
        }
        #endregion OnClosing

        #region OnClosed
        protected override void OnClosed(EventArgs e)
        {
            base.OnClosed(e);
            this.Dispose();
        }
        #endregion OnClosed

        #region Dispose
        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        /// <param key="disposing">true if managed resources should be disposed; otherwise, false.</param>
        protected override void Dispose(bool disposing)
        {
            if ( disposing )
            {
                if ( this.toolTip != null )
                {
                    this.toolTip.Dispose();
                    this.toolTip = null;
                }

                TextHighlighter highlighter = this.TextHighlighter;

                if ( highlighter != null )
                    highlighter.Dispose();

                if ( components != null )
                    components.Dispose();
            }

            base.Dispose(disposing);
        }
        #endregion Dispose

        #region GetCurrentWord
        internal string GetCurrentWord( int selStart, out int start )
        {
            start = selStart;
            string text = this.txtCode.Text;
            Match m = null;
            string s = text;

            //  Get the starting character position
            if ( text.Length == 0 )
                return string.Empty;
            if ( selStart < 1 )
                start = 0;
            else
            {
                s = text.Substring(0, selStart);
                m = this.regexBackToDelimiter.Match( s );
                if ( m.Success )
                {
                    start = m.Index + m.Length;
                    s = text.Substring( start );
                }
            }

            m = this.regexIdentifier.Match( s );
            return m.Success ? m.Value.Trim() : string.Empty;
        }
        #endregion GetCurrentWord

    }
}