Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Data
Imports System.Drawing
Imports System.Text
Imports System.Windows.Forms
Imports Infragistics.SyntaxParsing
Imports Infragistics.SyntaxParsing.Importing
Imports System.Drawing.Drawing2D

Public Partial Class frmMain
	Inherits Form
	#Region "Form Constructor"
	Public Sub New()
		InitializeComponent()
	End Sub
	#End Region

	#Region "Events"
    Private Sub frmMain_Load(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
        ' Handle the two custom events for cursor position and parse complete
        AddHandler txtEditor.CusorPositionChanged, New RichSyntaxEditor.CursorPositionChangedEventHandler(AddressOf txtEditor_CusorPositionChanged)
        AddHandler txtEditor.ParseComplete, New RichSyntaxEditor.ParseCompleteEventHandler(AddressOf txtEditor_ParseComplete)

        ' Add terminal symbol token names and colors for future parsing and colorization
        txtEditor.ColorizedTerminalSymbolTokens.Add("select", Color.Blue)
        txtEditor.ColorizedTerminalSymbolTokens.Add("from", Color.Blue)
        txtEditor.ColorizedTerminalSymbolTokens.Add("where", Color.Blue)
        txtEditor.ColorizedTerminalSymbolTokens.Add("limit", Color.Blue)
        txtEditor.ColorizedTerminalSymbolTokens.Add("or", Color.Blue)
        txtEditor.ColorizedTerminalSymbolTokens.Add("and", Color.Blue)

        ' Display the grammar definition in a read only text box
        txtEBNFDefinition.Text = My.Resources.SQL_EBNF

        ' Import grammar definition
        txtEditor.ImportGrammar(My.Resources.SQL_EBNF)

        ' Force initial parsing
        txtEditor.ParseEditor()
    End Sub

    Private Sub txtEditor_ParseComplete(ByVal output As String)
        lblRealTimeParsing.Text = output
    End Sub

    Private Sub txtEditor_CusorPositionChanged(ByVal output As String)
        lblCursorPosition.Text = output
    End Sub
#End Region
End Class

#Region "RichSyntaxEditor (generic inherited control for syntax coloring)"
Public Class RichSyntaxEditor
    Inherits RichTextBox
#Region "Variables"
    Private gmar As New Infragistics.SyntaxParsing.Grammar()
    Private Const WM_PAINT As Integer = &HF

    Private _ColorizedTerminalSymbolTokens As Dictionary(Of String, Color) = New Dictionary(Of [String], Color)()
    ''' <summary>
    ''' Dictionary of terminal symbol tokens and their font color.
    ''' </summary>
    Public Property ColorizedTerminalSymbolTokens() As Dictionary(Of String, Color)
        Get
            Return _ColorizedTerminalSymbolTokens
        End Get
        Set(ByVal value As Dictionary(Of String, Color))
            _ColorizedTerminalSymbolTokens = Value
        End Set
    End Property

    Private _isPainted As [Boolean] = True
    ''' <summary>
    ''' Determines whether or not to paint the editor. During the colorization process, you want this set to false.
    ''' </summary>
    Public Property isPainted() As [Boolean]
        Get
            Return _isPainted
        End Get
        Set(ByVal value As [Boolean])
            _isPainted = Value
            If Value Then
                Me.Invalidate()
            End If
        End Set
    End Property

    ''' <summary>
    ''' Event used to output the parsing results.
    ''' </summary>
    ''' <param name="output"></param>
    Public Delegate Sub ParseCompleteEventHandler(ByVal output As [String])
    Public Event ParseComplete As ParseCompleteEventHandler

    ''' <summary>
    ''' Event used to output the current cursor position.
    ''' </summary>
    ''' <param name="output"></param>
    Public Delegate Sub CursorPositionChangedEventHandler(ByVal output As [String])
    Public Event CusorPositionChanged As CursorPositionChangedEventHandler
#End Region

#Region "Methods"
    Public Sub ImportGrammar(ByVal GrammarDefinition As [String])
        ' Import the grammar definition from the application resources
        Dim ir As ImportResult = gmar.Import(GrammarDefinition, Infragistics.SyntaxParsing.GrammarDefinitionFormat.EBNF)
        ' Any errors? If so, display an error
        If ir.Errors.Count > 0 Then
            MessageBox.Show(My.Resources.ErrorImportMessage, My.Resources.ErrorImportTitle, MessageBoxButtons.OK, MessageBoxIcon.Warning)
        End If
    End Sub

    ''' <summary>
    ''' Forces the parsing of the editor.
    ''' </summary>
    Public Sub ParseEditor()
        Me.OnTextChanged(New EventArgs())
    End Sub

    Private Sub ProcessTokens(ByVal tokens As IEnumerable(Of ParseTreeTokenBase))
        For Each pttb As ParseTreeTokenBase In tokens
            If TypeOf pttb Is ErrorToken Then
                ' If token is an error token, color the text red
                Me.[Select](pttb.CharacterPosition, pttb.Length)
                Me.SelectionColor = Color.Red
                Continue For
            ElseIf pttb.TerminalSymbol Is Nothing OrElse ColorizedTerminalSymbolTokens.ContainsKey(pttb.TerminalSymbol.Value.ToLower()) = False Then
                ' If token isn't a terminal symbol and not an error, color it black
                Me.[Select](pttb.CharacterPosition, pttb.Length)
                If Me.SelectionColor <> Color.Black Then
                    Me.SelectionColor = Color.Black
                End If
                Continue For
            ElseIf ColorizedTerminalSymbolTokens.ContainsKey(pttb.TerminalSymbol.Value.ToLower()) = False Then
                ' If token is a defined terminal symbol, but isn't in our dictionary, color it black
                Me.[Select](pttb.CharacterPosition, pttb.Length)
                Me.SelectionColor = Color.Black
                Continue For
            End If

            ' Token is a terminal symbol and is found in dictionary,
            '   color it based on defined color
            Me.[Select](pttb.CharacterPosition, pttb.Length)
            Me.SelectionColor = CType(ColorizedTerminalSymbolTokens(pttb.TerminalSymbol.Value.ToLower()), Color)
        Next
    End Sub
#End Region

#Region "Processed Events"
    ''' <summary>
    ''' Override the paint method of the control to disallow painting while processing tokens.
    ''' </summary>
    ''' <param name="m"></param>
    Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)
        If m.Msg = WM_PAINT Then
            If isPainted Then
                MyBase.WndProc(m)
            Else
                m.Result = IntPtr.Zero
            End If
        Else
            MyBase.WndProc(m)
        End If
    End Sub

    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        ' Copy the current position of the cursor
        Dim insertionPointPosition As Integer = Me.SelectionStart

        ' Output the cursor position
        RaiseEvent CusorPositionChanged([String].Format(My.Resources.CursorPosition, Me.GetLineFromCharIndex(insertionPointPosition) + 1, (insertionPointPosition - Me.GetFirstCharIndexFromLine(Me.GetLineFromCharIndex(insertionPointPosition))) + 1))

        ' If the editor has no text, then return
        If Me.Text.Length = 0 Then
            Return
        End If

        Me.isPainted = False

        ' Parse the text editor
        Dim pt As ParseTree = gmar.Parse(Me.Text)

        ' Does the Nodes property contain nodes and no have errors?
        If pt.Nodes IsNot Nothing AndAlso pt.Errors.Count = 0 Then
            Dim startTime As DateTime = DateTime.Now

            ' Process the tokens for colorization
            ProcessTokens(pt.Tokens)

            Dim endTime As DateTime = DateTime.Now
            Dim duration As TimeSpan = endTime - startTime

            ' Output the parsing result with how long the parsing took
            RaiseEvent ParseComplete([String].Format(My.Resources.ParsingSuccess, duration))

            ' Force update cursor position for info
            Me.OnSelectionChanged(New EventArgs())
        Else
            ' There were errors, but process the tokens
            ProcessTokens(pt.Tokens)
            ' Output the parsing result with the error information
            RaiseEvent ParseComplete([String].Format(My.Resources.ParsingError, pt.Errors(0)))
        End If

        ' Restore the cursor to the correct location
        Me.[Select](insertionPointPosition, 0)

        ' Set the text color to black
        Me.SelectionColor = Color.Black

        Me.isPainted = True
    End Sub

    Protected Overrides Sub OnSelectionChanged(ByVal e As EventArgs)
        ' If the editor is currently parsing, return
        If Not isPainted Then
            Return
        End If

        ' Copy the current position of the cursor
        Dim insertionPointPosition As Integer = Me.SelectionStart

        ' Output the cursor position
        RaiseEvent CusorPositionChanged([String].Format(My.Resources.CursorPosition, Me.GetLineFromCharIndex(insertionPointPosition) + 1, (insertionPointPosition - Me.GetFirstCharIndexFromLine(Me.GetLineFromCharIndex(insertionPointPosition))) + 1))
    End Sub
#End Region
End Class
#End Region
